﻿using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.Data;
using System.Drawing;
using System.Linq;
using System.Text;
using System.Windows.Forms;

using System.IO;
using System.Drawing.Imaging;
using System.Drawing.Drawing2D;

namespace howto_image_spiral
{
    public partial class Form1 : Form
    {
        public Form1()
        {
            InitializeComponent();
        }

        private Bitmap BaseImage = Properties.Resources.cat;
        private Bitmap BgImage = null;
        private Color BgColor = Color.White;

        private void mnuFileOpen_Click(object sender, EventArgs e)
        {
            if (ofdImage.ShowDialog() == DialogResult.OK)
            {
                BaseImage = LoadBitmapUnlocked(ofdImage.FileName);
            }
        }

        private void mnuFileBackgroundColor_Click(object sender, EventArgs e)
        {
            cdBackgroundColor.Color = BgColor;
            if (cdBackgroundColor.ShowDialog() == DialogResult.OK)
            {
                BgColor = cdBackgroundColor.Color;
                Bitmap bm = new Bitmap(16, 16);
                using (Graphics gr = Graphics.FromImage(bm))
                {
                    gr.Clear(BgColor);
                    gr.DrawRectangle(Pens.Black, 0, 0, 15, 15);
                }
                mnuFileBackgroundColor.Image = bm;
                BgImage = null;
            }
        }

        private void mnuFileBackgroundImage_Click(object sender, EventArgs e)
        {
            if (ofdImage.ShowDialog() == DialogResult.OK)
            {
                BgImage = LoadBitmapUnlocked(ofdImage.FileName);
                txtWidth.Text = BgImage.Width.ToString();
                txtHeight.Text = BgImage.Height.ToString();
            }
        }

        private void mnuFileSaveAs_Click(object sender, EventArgs e)
        {
            if (sfdImage.ShowDialog() == DialogResult.OK)
            {
                try
                {
                    SaveImage(picResult.Image, sfdImage.FileName);
                }
                catch (Exception ex)
                {
                    MessageBox.Show(ex.Message);
                }
            }
        }

        private void mnuFileExit_Click(object sender, EventArgs e)
        {
            Close();
        }

        // If we have a background and base image, make the spiral.
        private void MakeSpiral()
        {
            // Get the parameters.
            int width, height, num_spirals;
            float scale, A, B, dtheta;
            if (ParametersInvalid(out width, out height,
                out scale, out A, out B, out dtheta, out num_spirals)) return;

            // Size the result PictureBox and the form.
            picResult.ClientSize = new Size(width, height);
            int wid = Math.Min(panScroll.Left + picResult.Width + panScroll.Left, 800);
            int hgt = Math.Min(panScroll.Top + picResult.Width + panScroll.Left, 800);
            if (wid < ClientSize.Width) wid = ClientSize.Width;
            if (hgt < ClientSize.Height) hgt = ClientSize.Height;
            ClientSize = new Size(wid, hgt);

            // Make the result bitmap.
            Bitmap bm = new Bitmap(width, height);
            using (Graphics gr = Graphics.FromImage(bm))
            {
                gr.SmoothingMode = SmoothingMode.AntiAlias;
                if (BgImage != null)
                {
                    // Draw the background image.
                    using (Brush brush = new TextureBrush(BgImage))
                    {
                        gr.FillRectangle(brush, 0, 0, width, height);
                    }
                }
                else
                {
                    // Fill the background with the background color.
                    gr.Clear(BgColor);
                }

                // Draw the spiral.
                DrawImageSpiral(gr, BaseImage, width, height,
                    scale, A, B, dtheta, num_spirals);
            }

            // Display the result.
            picResult.Image = bm;
        }

        // Get the parameters. If any of the parameters is
        // missing or invalid, tell the user and return true.
        private bool ParametersInvalid(
            out int width, out int height,
            out float scale, out float A, out float B,
            out float dtheta, out int num_spirals)
        {
            width = 0;
            height = 0;
            scale = 0;
            A = 0;
            B = 0;
            num_spirals = 0;
            dtheta = 0;

            if (!int.TryParse(txtWidth.Text, out width) || (width <= 0))
            {
                MessageBox.Show("Please enter a width greater than zero.");
                txtWidth.Focus();
                return true;
            }

            if (!int.TryParse(txtHeight.Text, out height) || (height <= 0))
            {
                MessageBox.Show("Please enter a height greater than zero.");
                txtHeight.Focus();
                return true;
            }

            if (!float.TryParse(txtScale.Text, out scale) ||
                (Math.Abs(scale) <= 0.01))
            {
                MessageBox.Show("Please enter a non-zero scale.");
                txtScale.Focus();
                return true;
            }

            // Make sure we have a base image.
            if (BaseImage == null)
            {
                MessageBox.Show("Please select a base image.");
                return true;
            }

            // Get the spiral parameters.
            if (!float.TryParse(txtA.Text, out A))
            {
                MessageBox.Show("Please enter parameter A.");
                return true;
            }

            if (!float.TryParse(txtB.Text, out B))
            {
                MessageBox.Show("Please enter parameter B.");
                return true;
            }

            if (!float.TryParse(txtDtheta.Text, out dtheta))
            {
                MessageBox.Show("Please enter parameter DTheta.");
                return true;
            }
            // Convert from degrees to radians.
            dtheta *= (float)(Math.PI / 180);

            if (!int.TryParse(txtNumSpirals.Text, out num_spirals) || num_spirals <= 0)
            {
                MessageBox.Show("Please enter a number if spirals greater than zero.");
                return true;
            }

            return false;
        }

        // Draw the spiral of images.
        private void DrawImageSpiral(Graphics gr, Bitmap image,
            int width, int height, float scale, float A, float B,
            float dtheta, int num_spirals)
        {
            // Find the maximum distance to the rectangle's corners.
            PointF center = new PointF(width / 2, height / 2);
            float max_r = (float)Math.Sqrt(
                center.X * center.X + center.Y * center.Y) +
                BaseImage.Height;

            // Draw the spirals.
            float start_angle = 0;
            float d_start = (float)(2 * Math.PI / num_spirals);
            for (int i = 0; i < num_spirals; i++)
            {
                List<PointF> points =
                    GetSpiralPoints(center, A, B, start_angle, dtheta, max_r);

                foreach (PointF point in points)
                {
                    float dist = Distance(center, point);
                    float r = dist / 2;
                    DrawImageAt(gr, center, point, r, scale);
                }

                start_angle += d_start;
            }
        }

        private float Distance(PointF point1, PointF point2)
        {
            float dx = point1.X - point2.X;
            float dy = point1.Y - point2.Y;
            return (float)Math.Sqrt(dx * dx + dy * dy);
        }

        // Draw the base image at the indicated point
        // appropriately scaled and rotated.
        private void DrawImageAt(Graphics gr, PointF center,
            PointF point, float r, float image_scale)
        {
            float dx = point.X - center.X;
            float dy = point.Y - center.Y;
            float angle = (float)(Math.Atan2(dy, dx) * 180 / Math.PI) + 90;
            float sx = r / BaseImage.Width * image_scale;
            float sy = r / BaseImage.Height * image_scale;
            float scale = Math.Min(sx, sy);
            GraphicsState state = gr.Save();
            gr.ResetTransform();
            gr.ScaleTransform(scale, scale);
            gr.RotateTransform(angle, MatrixOrder.Append);
            gr.TranslateTransform(point.X, point.Y, MatrixOrder.Append);

            PointF[] dest_points =
            {
                new PointF(-BaseImage.Width / 2, -BaseImage.Height / 2),
                new PointF(BaseImage.Width / 2, -BaseImage.Height / 2),
                new PointF(-BaseImage.Width / 2, BaseImage.Height / 2),
            };
            RectangleF src_rect = new RectangleF(0, 0, BaseImage.Width, BaseImage.Height);

            if (chkDarken.Checked)
            {
                float radius = Math.Min(center.X, center.Y) / 3f;
                float dark_scale = r / radius + 0.2f;
                if (dark_scale > 1) dark_scale = 1;
                Bitmap bm = AdjustBrightness(BaseImage, dark_scale);
                gr.DrawImage(bm, dest_points, src_rect, GraphicsUnit.Pixel);
            }
            else
            {
                gr.DrawImage(BaseImage, dest_points, src_rect, GraphicsUnit.Pixel);
            }

            gr.Restore(state);
        }


        // Return points that define a spiral.
        // See http://csharphelper.com/blog/2018/11/draw-a-logarithmic-spiral-in-c/
        private List<PointF> GetSpiralPoints(
            PointF center, float A, float B,
            float angle_offset, float dtheta, float max_r)
        {
            // Get the points.
            List<PointF> points = new List<PointF>();
            float min_theta = (float)(Math.Log(0.1 / A) / B);
            for (float theta = min_theta; ; theta += dtheta)
            {
                // Calculate r.
                float r = (float)(A * Math.Exp(B * theta));

                // Convert to Cartesian coordinates.
                float x, y;
                PolarToCartesian(r, theta + angle_offset, out x, out y);

                // Center.
                x += center.X;
                y += center.Y;

                // Create the point.
                points.Add(new PointF((float)x, (float)y));

                // If we have gone far enough, stop.
                if (r > max_r) break;
            }
            return points;
        }

        private void PolarToCartesian(float r, float theta, out float x, out float y)
        {
            x = (float)(r * Math.Cos(theta));
            y = (float)(r * Math.Sin(theta));
        }

        // Load a bitmap without locking it.
        private Bitmap LoadBitmapUnlocked(string file_name)
        {
            try
            {
                using (Bitmap bm = new Bitmap(file_name))
                {
                    return new Bitmap(bm);
                }
            }
            catch (Exception ex)
            {
                MessageBox.Show(ex.Message);
                return null;
            }
        }

        // Save the file with the appropriate format.
        public void SaveImage(Image image, string filename)
        {
            try
            {
                string extension = Path.GetExtension(filename);
                switch (extension.ToLower())
                {
                    case ".bmp":
                        image.Save(filename, ImageFormat.Bmp);
                        break;
                    case ".exif":
                        image.Save(filename, ImageFormat.Exif);
                        break;
                    case ".gif":
                        image.Save(filename, ImageFormat.Gif);
                        break;
                    case ".jpg":
                    case ".jpeg":
                        image.Save(filename, ImageFormat.Jpeg);
                        break;
                    case ".png":
                        image.Save(filename, ImageFormat.Png);
                        break;
                    case ".tif":
                    case ".tiff":
                        image.Save(filename, ImageFormat.Tiff);
                        break;
                    default:
                        throw new NotSupportedException(
                            "Unknown file extension " + extension);
                }
            }
            catch (Exception ex)
            {
                MessageBox.Show(ex.Message);
            }
        }

        private void btnDraw_Click(object sender, EventArgs e)
        {
            Cursor = Cursors.WaitCursor;
            MakeSpiral();
            Cursor = Cursors.Default;
        }

        // Adjust the image's brightness.
        // See http://csharphelper.com/blog/2014/10/use-an-imageattributes-object-to-adjust-an-images-brightness-in-c/
        private Bitmap AdjustBrightness(Image image, float brightness)
        {
            // Make the ColorMatrix.
            float b = brightness;
            ColorMatrix cm = new ColorMatrix(new float[][]
            {
                new float[] {b, 0, 0, 0, 0},
                new float[] {0, b, 0, 0, 0},
                new float[] {0, 0, b, 0, 0},
                new float[] {0, 0, 0, 1, 0},
                new float[] {0, 0, 0, 0, 1},
            });
            ImageAttributes attributes = new ImageAttributes();
            attributes.SetColorMatrix(cm);

            // Draw the image onto the new bitmap while applying
            // the new ColorMatrix.
            Point[] points =
            {
                new Point(0, 0),
                new Point(image.Width, 0),
                new Point(0, image.Height),
            };
            Rectangle rect = new Rectangle(0, 0, image.Width, image.Height);

            // Make the result bitmap.
            Bitmap bm = new Bitmap(image.Width, image.Height);
            using (Graphics gr = Graphics.FromImage(bm))
            {
                gr.DrawImage(image, points, rect,
                    GraphicsUnit.Pixel, attributes);
            }

            // Return the result.
            return bm;
        }
    }
}
